@@ -87,7 +87,9 @@ L<Mail::DKIM::Policy/"apply()">.
 
 =head1 CONSTRUCTOR
 
-=head2 new() - construct an object-oriented verifier
+=head2 new()
+
+Constructs an object-oriented verifier.
 
   my $dkim = Mail::DKIM::Verifier->new();
 
@@ -109,7 +111,8 @@ is written to the referenced string or file handle.
 package Mail::DKIM::Verifier;
 use base "Mail::DKIM::Common";
 use Carp;
-our $VERSION = 0.35;
+our $VERSION = 0.40;
+our $MAX_SIGNATURES_TO_PROCESS = 50;
 
 sub init
 {
@@ -155,6 +158,7 @@ sub handle_header
 		{
 			my $signature = Mail::DKIM::Signature->parse($line);
 			$self->add_signature($signature);
+			$signature->fetch_public_key;
 		};
 		if ($@)
 		{
@@ -174,6 +178,7 @@ sub handle_header
 		{
 			my $signature = Mail::DKIM::DkSignature->parse($line);
 			$self->add_signature($signature);
+			$signature->fetch_public_key;
 		};
 		if ($@)
 		{
@@ -194,6 +199,10 @@ sub add_signature
 	croak "wrong number of arguments" unless (@_ == 1);
 	my ($signature) = @_;
 
+	# ignore signature headers once we've seen 50 or so
+	# this protects against abuse.
+	return if (@{$self->{signatures}} > $MAX_SIGNATURES_TO_PROCESS);
+
 	push @{$self->{signatures}}, $signature;
 
 	unless ($self->check_signature($signature))
@@ -287,16 +296,21 @@ sub check_signature
 		return 0;
 	}
 
-	unless ($signature->domain ne '')
+	unless (defined $signature->domain)
 	{
 		# no domain specified
-		$self->{signature_reject_reason} =
-			!defined($signature->domain) ? "missing d tag"
-			: "invalid domain in d tag";
+		$self->{signature_reject_reason} = "missing d tag";
 		return 0;
 	}
 
-	unless ($signature->selector)
+	if ($signature->domain eq '')
+	{
+		# blank domain
+		$self->{signature_reject_reason} = "invalid domain in d tag";
+		return 0;
+	}
+
+	unless (defined $signature->selector)
 	{
 		# no selector specified
 		$self->{signature_reject_reason} = "missing s tag";
@@ -389,45 +403,7 @@ sub finish_header
 
 	foreach my $algorithm (@{$self->{algorithms}})
 	{
-		$algorithm->finish_header;
-	}
-
-	# For each parsed signature, check it for validity. If none are valid,
-	# our result is "invalid" and our result detail will be the reason
-	# why the last signature was invalid.
-
-	foreach my $signature (@{$self->{signatures}})
-	{
-		unless (check_signature_identity($signature))
-		{
-			$self->{signature_reject_reason} = "bad identity";
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
-		}
-
-		# get public key
-		my $pkey;
-		eval
-		{
-			$pkey = $signature->get_public_key;
-		};
-		if ($@)
-		{
-			my $E = $@;
-			chomp $E;
-			$self->{signature_reject_reason} = "public key: $E";
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
-		}
-
-		unless ($self->check_public_key($signature, $pkey))
-		{
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
-		}
+		$algorithm->finish_header(Headers => $self->{headers});
 	}
 
 	# stop processing signatures that are already known to be invalid
@@ -446,14 +422,37 @@ sub finish_header
 	}
 }
 
-sub finish_body
+sub _check_and_verify_signature
 {
 	my $self = shift;
+	my ($algorithm) = @_;
 
-	foreach my $algorithm (@{$self->{algorithms}})
-	{
-		# finish canonicalizing
-		$algorithm->finish_body;
+		# check signature
+		my $signature = $algorithm->signature;
+		unless (check_signature_identity($signature))
+		{
+			$self->{signature_reject_reason} = "bad identity";
+			return ("invalid", $self->{signature_reject_reason});
+		}
+
+		# get public key
+		my $pkey;
+		eval
+		{
+			$pkey = $signature->get_public_key;
+		};
+		if ($@)
+		{
+			my $E = $@;
+			chomp $E;
+			$self->{signature_reject_reason} = "public key: $E";
+			return ("invalid", $self->{signature_reject_reason});
+		}
+
+		unless ($self->check_public_key($signature, $pkey))
+		{
+			return ("invalid", $self->{signature_reject_reason});
+		}
 
 		# verify signature
 		my $result;
@@ -479,6 +478,19 @@ sub finish_body
 			$result = "fail";
 			$details = $E;
 		}
+	return ($result, $details);
+}
+
+sub finish_body
+{
+	my $self = shift;
+
+	foreach my $algorithm (@{$self->{algorithms}})
+	{
+		# finish canonicalizing
+		$algorithm->finish_body;
+
+		my ($result, $details) = $self->_check_and_verify_signature($algorithm);
 
 		# save the results of this signature verification
 		$algorithm->{result} = $result;
@@ -498,7 +510,9 @@ sub finish_body
 
 =head1 METHODS
 
-=head2 PRINT() - feed part of the message to the verifier
+=head2 PRINT()
+
+Feeds part of the message to the verifier.
 
   $dkim->PRINT("a line of the message\015\012");
   $dkim->PRINT("more of");
@@ -508,14 +522,24 @@ Feeds content of the message being verified into the verifier.
 The API is designed this way so that the entire message does NOT need
 to be read into memory at once.
 
-=head2 CLOSE() - call this when finished feeding in the message
+Please note that although the PRINT() method expects you to use
+SMTP-style line termination characters, you should NOT use the
+SMTP-style dot-stuffing technique described in RFC 2821 section 4.5.2.
+Nor should you use a <CR><LF>.<CR><LF> sequence to terminate the
+message.
+
+=head2 CLOSE()
+
+Call this when finished feeding in the message.
 
   $dkim->CLOSE;
 
 This method finishes the canonicalization process, computes a hash,
 and verifies the signature.
 
-=head2 fetch_author_domain_policies() - retrieves ADSP records from DNS
+=head2 fetch_author_domain_policies()
+
+Retrieves ADSP records from DNS.
 
   my @policies = $dkim->fetch_author_domain_policies;
   foreach my $policy (@policies)
@@ -542,7 +566,8 @@ sub fetch_author_domain_policies
 	my $self = shift;
 	use Mail::DKIM::AuthorDomainPolicy;
 
-	return () unless $self->{headers_by_name}->{from};
+	return () unless ($self->{headers_by_name}->{from} &&
+	    $self->{headers_by_name}->{from} =~ /\@/);
 	my @list = Mail::Address->parse(
 		$self->{headers_by_name}->{from}
 		);
@@ -550,14 +575,16 @@ sub fetch_author_domain_policies
 
 	# fetch the policies
 	return map {
-		Mail::DKIM::DkimPolicy->fetch(
+		Mail::DKIM::AuthorDomainPolicy->fetch(
 			Protocol => "dns",
 			Author => $_,
 			)
 		} @authors;
 }
 
-=head2 fetch_author_policy() - retrieves a signing policy from DNS
+=head2 fetch_author_policy()
+
+Retrieves a signing policy from DNS.
 
   my $policy = $dkim->fetch_author_policy;
   my $policy_result = $policy->apply($dkim);
@@ -587,7 +614,9 @@ sub fetch_author_policy
 			);
 }
 
-=head2 fetch_sender_policy() - retrieves a signing policy from DNS
+=head2 fetch_sender_policy()
+
+Retrieves a signing policy from DNS.
 
   my $policy = $dkim->fetch_sender_policy;
   my $policy_result = $policy->apply($dkim);
@@ -626,7 +655,9 @@ sub fetch_sender_policy
 			);
 }
 
-=head2 load() - load the entire message from a file handle
+=head2 load()
+
+Load the entire message from a file handle.
 
   $dkim->load($file_handle);
 
@@ -634,7 +665,9 @@ Reads a complete message from the designated file handle,
 feeding it into the verifier. The message must use <CRLF> line
 terminators (same as the SMTP protocol).
 
-=head2 message_originator() - access the "From" header
+=head2 message_originator()
+
+Access the "From" header.
 
   my $address = $dkim->message_originator;
 
@@ -650,7 +683,9 @@ To get just the email address part, do:
 
 See also L</"message_sender()">.
 
-=head2 message_sender() - access the "From" or "Sender" header
+=head2 message_sender()
+
+Access the "From" or "Sender" header.
 
   my $address = $dkim->message_sender;
 
@@ -669,7 +704,9 @@ transmission of the message. For example, if a secretary were to send a
 message for another person, the "sender" would be the secretary and
 the "originator" would be the actual author.
 
-=head2 policies() - retrieves applicable signing policies from DNS
+=head2 policies()
+
+Retrieves applicable signing policies from DNS.
 
   my @policies = $dkim->policies;
   foreach my $policy (@policies)
@@ -708,7 +745,9 @@ sub policies
 }
 
 
-=head2 result() - access the result of the verification
+=head2 result()
+
+Access the result of the verification.
 
   my $result = $dkim->result;
 
@@ -747,13 +786,15 @@ Returned if no DKIM-Signature headers (valid or invalid) were found.
 
 In case of multiple signatures, the "best" result will be returned.
 Best is defined as "pass", followed by "fail", "invalid", and "none".
-To examine the results of individual signatures, use the L</signatures>
+To examine the results of individual signatures, use the L</"signatures()">
 method to retrieve the signature objects. See
-L<Mail::DKIM::Signature/result>.
+L<Mail::DKIM::Signature/"result()">.
 
 =cut
 
-=head2 result_detail() - access the result, plus details if available
+=head2 result_detail()
+
+Access the result, plus details if available.
 
   my $detail = $dkim->result_detail;
 
@@ -794,7 +835,9 @@ The following are possible results from the result_detail() method:
   invalid (public key: OpenSSL error: ...)
   none
 
-=head2 signature() - access the message's DKIM signature
+=head2 signature()
+
+Access the message's DKIM signature.
 
   my $sig = $dkim->signature;
 
@@ -807,7 +850,9 @@ Best is defined as "pass", followed by "fail", "invalid", and "none".
 
 =cut
 
-=head2 signatures() - access all of this message's signatures
+=head2 signatures()
+
+Access all of this message's signatures.
 
   my @all_signatures = $dkim->signatures;
 
